/* pcnethme.h: An AMD PCnet32 ethernet driver header for linux. */
/*
 *  Copyright 1996-1999 Thomas Bogendoerfer
 * 
 *      Derived from the lance driver written 1993,1994,1995 by Donald Becker
 *
 * 	Copyright 1993 United States Government as represented by the
 * 	Director, National Security Agency.
 * 
 * 	This software may be used and distributed according to the terms
 * 	of the GNU Public License, incorporated herein by reference.
 *
 * 	This driver is for PCnet32 and PCnetPCI based homenetwork cards.
 */


#include <linux/config.h>
#include <linux/module.h>
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/dma.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>


#ifdef HOMENET

#define     TRUE        1
#define     FALSE       0

#define     TIMER_SPEED 2           // two seconds 

#define	ETHPHY_AUTO					0x7f
#define	ETHPHY_10HD					0x01
#define	ETHPHY_10FD					0x02
#define	ETHPHY_100HD				0x04
#define	ETHPHY_100FD				0x08

/*
 * INTPHY VALUES 
 */ 

#define	INTPHY_AUTO					ETHPHY_AUTO
#define	INTPHY_10HD					ETHPHY_10HD
#define	INTPHY_10FD					ETHPHY_10FD

/* Values also match hardware bit defs unless indicated. */

#define	PHONELINE_AUTO				0x0	      /* does not match HW */
#define	PHONELINE_AUTO_HIGH			0x1	    /* does not match HW */
#define	PHONELINE_LPLS				0xA
#define	PHONELINE_LPHS				0x9
#define	PHONELINE_HPLS				0x6
#define	PHONELINE_HPHS				0x5


#define	PHYSEL_SFEX				0
#define	PHYSEL_HOMEPNA			1
#define	PHYSEL_EXT				2
#define	PHYSEL_AUTO				3



#define	EXTPHY_AUTO					ETHPHY_AUTO
#define	EXTPHY_10HD					ETHPHY_10HD
#define	EXTPHY_10FD					ETHPHY_10FD
#define	EXTPHY_100HD				ETHPHY_100HD
#define	EXTPHY_100FD				ETHPHY_100FD

/*
 *	Phy modes to indicate auto or manual Phy mode
 */
#define	PHY_SEL_AUTO				0
#define	PHY_SEL_MANUAL				1

#define PHY_SEL_SFEX			0x0000
#define PHY_SEL_HOMERUN			0x0001
#define PHY_SEL_EXTERNAL		0x0002
#define FIND_ACTIVE_PHY			0x0003

/*
 * Maximum Phy supported by HomeLan
 */
#define MAX_EXT_PHYS             3
#define MAX_PHYS             	 3
#define MAX_NEIGHBOR_STORAGE    10      /* cannot be greater than 16 */
#define ALLOWED_CRC_PER_SEC      5


/* Define BCR49 : New register in Everest */
#define LANCE_BCR49				49
#define PHY_SEL_LOCK				0x0080
#define PHY_SEL_MASK				0xFFFC
#define PHY_SEL_DEFAULT_MASK	0x0300
#define PHY_SEL_DEFAULT_BIT	8
 
/* private->reset_phy_status state values */

#define	RSTPHY_INIT         1
#define	RSTPHY_WAIT         2
#define	RSTPHY_COMPLETE     3


/* Test Register 28 (IADJ) defs */
#define	REG_TST28				28
#define	IADJ_MASK				0x0F00
#define	REG28_DEFAULT			0x0500

struct phy_desc
{
	u16	phy_speed_set;
	u16	phy_address;
	u16	phy_mfr_id;
	u16	phy_ieee_addr;
	u16	phy_mode_set;
	u16	phy_ability;
	u16	phy_link_set;
	int	phy_exists;

}; 

struct auto_config {
	u32   stateHistory;
	u32   MaxAllowedCRCs;
	u32   numberOfCRCs;
	u32   numberOfExcludeSources;
	u8   excludeSources[MAX_NEIGHBOR_STORAGE][ETH_ALEN];
	u8   CRCSources[MAX_NEIGHBOR_STORAGE][ETH_ALEN];
	u32   CRCStates;
	u32   RcvStateValid;
}; 

#define INIT_STATE_MASK 0xC0000000
#define STATE_BIT_MASK  0x00000003
#define STATE_BITS      2
#define INIT_STATE_BITS 30

#define STATE_LPHS      0x0001
#define STATE_HPHS      0x0002
#define STATE_HPLS      0x0003
#define STATE_LPLS      0x0004
#define	STATE_RX		0xFFFF
#define	STATE_MANUAL	0xEFFF

#define HOMELAN_TIMEOUT 1000            // 1 second

#define PCNET_HOME_A1                   0x0001
#define PCNET_HOME_B0                   0x0002
#define PCNET_HOME_B1                   0x0003

/* revision IDs */

#define    PCNET_HOME_A1_REV_ID   0x50
#define    PCNET_HOME_B0_REV_ID   0x51
#define    PCNET_HOME_B1_REV_ID   0x52


#define ANR0        0x0000

/* ANR0 (common for all PHYs) */
#define ANR0_PR     0x8000          // PHY Reset: 1 (external PHY).
#define ANR0_LB     0x4000          // Loopback: 1 = Within external PHY.
#define ANR0_SS     0x2000          // Speed Selection: 1 = 100 Mb/s.
#define ANR0_ANE    0x1000          // Auto-Negotiation Enable: 0 = disabled.
#define ANR0_DM     0x0100          // Duplex Mode: 1 = Full Duplex.
#define	ANR0_RA		0x0200			// Restart auto-negotiation
#define ANR0_ISO	0x0400			// Isolate Phy from MII

//
//	Definition for the Home Lan.
//

/* HomeRun Phy Control Register */
#define	HR_PHY_CTRL_REG	0x10
#define	MASTER_CMD			0x8000
#define	C_LOW_POWER			0x0800
#define	C_HIGH_POWER		0x0400
#define	C_LOW_SPEED			0x0200
#define	C_HIGH_SPEED		0x0100
#define	HIGH_SPEED			0x0004
#define	HIGH_POWER			0x0002
#define  HIGH_POWER_BIT		1
#define	AID_COMMAND_MASK	0x90FF
#define	LOCAL_CMD_MASK		0x9FF0

/* Phy Status Register */
#define PHY_STATUS_REG			0x11
#define	LINK_PULSE_DISABLE		0x1000
#define	RECEIVED_POWER			0x0040
#define	RECEIVED_SPEED			0x0020
#define	RECEIVED_SPEED_BIT 		5

/* HomePNA PHY Event Status Register */
#define HR_PHY_EVENT_STAT_REG	26
#define	PHYEVENT_PKTRX			0x08	/* Packet Received of any type */
#define	PHYEVENT_PKTTX			0x04	/* Packet Transmitted of any type */
#define	PHYEVENT_RCRX			0x02	/* Valid Remote Command received */
#define	PHYEVENT_RCTX			0x01	/* Remote Command Sent */

#define HPNA_PHYNOISE_REG		23
#define HPNA_NOISE_DEFAULTS		0x0330

#define	HPNA_PHYSTATS_REG		25

/* HomeRun Analog Test Register 0x1F */
#define 	HR_ANALOG_TEST_REG		0x1F
#define 	FORCE_LINK_VALID			0x0080

/* HomeRun Drive Level Control */
#define 	DEFAULT_LEVEL_CTRL	0x1589


#define	PHYADD_EXTERN			0x02

#define	PHYADD_HOMEPNA			0x1D
#define	PHYADD_SFEX				0x1E

/* HomeLan PHY Address */
#define HOMEPNA_PHY_ADDRESS	(PHYADD_HOMEPNA << 5)
#define SFEX_PHY_ADDRESS	(PHYADD_SFEX	<< 5)	//Default Phy for HomeLan, same as PHYADDR in lancehrd.h
#define EXT_PHY_END_POINT		0x1D
#define	NUM_BUILTIN_PHYS		2
#define PHY_ADDRESS_MASK		(0x1F << 5)

/* Address of internal PHY, aligned starting at bit 5 */
#define	PHYADDR			(0x01 << 5)

#define	DANAS			0x0080	/* Disable Auto-Negotiation auto setup*/
#define	XPHYRST			0x0040	/* External PHY Reset */
#define	XPHYANE			0x0020	/* External PHY Auto-Negotiation Enable */
#define	XPHYFD			0x0010	/* External PHY Full-Duplex Enable */
#define	XPHYSP			0x0008	/* External PHY Speed Select (1=100Mbps) */
#define	MIIuL			0x0004	/* Micro Linear 6692 PHY Select */
#define	MIIILP			0x0002	/* Internal Loopback Enable (MII) */
#define	FCON			0x0001	/* Fast Config Enable */
#define	FORCED_PHY_MASK	0xFF07
/* CSR Registers */

/* Defs for reading link status from [MII */
/* Used in function LanceReadLink() */
#define	MII_LINKSTAT			0x0004	/* Link Status Bit */
#define	MII_AUTONEG_ABLE		0x0008	/* Phy is able to auto-negotiate */
#define	MII_AUTONEG_COMP		0x0020	/* Auto-negotiation complete */
#define	MII_STAT_REG			0x0001	/* Offset of status register */
#define	MII_CTRL_REG			0x0000	/* Offset of control register */

// for Homelan
#define	MII_IEEE_ID				0x0002	/* Offset of IEEE ID register */
#define	MII_IEEE_ID1			0x0003	/* Offset of IEEE ID register */
#define	MII_MY_ABILITY			0x0004	/* Auto-negotiation advertisement ability register */
#define	ABILITY_802_3			0x0001	/* Bit to test 802.3 ability of an external tranceiver */

#define	MII_LNK_PRTNR			0x0005	/* Auto-negotiation link partner ability register */
#define	ABILITY_MASK			0x1FC0	/* Mask to isolate the technology ability field */
#define	MII_ABILITY_MASK		0x01E0	/* Mask to isolate 10/100 & HD/FD capabilities */
#define	MII_TECH_MASK			0x001F	/* Mask to isolate 802.3 compliance data */
#define	MII_ADDR				33		/* AKA BCR33, used to access tranceiver */
#define	MII_MDR					34		/* AKA BCR34, used to read data from tranceiver */
#define	BCR33_SHADOW			0x8000	
#define	LANCE_BCR33	 			33
#define	LANCE_BCR34	 			34

#define IPGMASK                 0xff
#define IPGVAL                  0x64

#define MAC_HEADER_SIZE (UINT)14

/* kernel related defines */
#define RUN_AT(x) (jiffies + (x))

static int     check_phy_status (struct device *dev);
static void    disable_ext_phy(struct device *dev, u32 phy_num);
static void    enable_ext_phy(struct device *dev, u32 phy_num);
static void    isolate_phys(struct device *dev);
static void    home_link_monitor(struct device *dev);
static void    reset_phy( struct device *dev);
static void    set_network_state(struct device *dev, u16 state);
static void    set_local_state(struct device *dev, u16 state);
static void    homelan_monitor(struct device *dev);
static u16     homelan_sync_rate(struct device *dev);
static void    phy_setting(struct device *dev);
static void    everest_bug_fix(struct device *dev);
static void    link_pulse_control(struct device *dev);
static int     phy_search(struct device *dev);
static void    init_phy_info(struct device *dev);
static void    init_active_phy (struct device *dev); 
static void    init_ethernet_phy(struct device *dev); 
static void    init_home_phy(struct device *dev);
static int     find_ext_phy(struct device *dev);
static int     get_phy_link_state(struct device *dev, int phy_num);
static void    crc_error_handler(struct device *dev, int status);

#endif /* ifdef homnet */

/*---------------- End of HomeNet related -----------------------------*/

#define PORT_AUI      0x00
#define PORT_10BT     0x01
#define PORT_GPSI     0x02
#define PORT_MII      0x03

#define PORT_PORTSEL  0x03
#define PORT_ASEL     0x04
#define PORT_100      0x40
#define PORT_FD       0x80



/*
 * Set the number of Tx and Rx buffers, using Log_2(# buffers).
 * Reasonable default values are 4 Tx buffers, and 16 Rx buffers.
 * That translates to 2 (4 == 2^^2) and 4 (16 == 2^^4).
 */

#ifndef PCNET32_LOG_TX_BUFFERS
#define PCNET32_LOG_TX_BUFFERS 4
#define PCNET32_LOG_RX_BUFFERS 4
#endif

#define MAX_UNITS   8

#define TX_RING_SIZE			(1 << (PCNET32_LOG_TX_BUFFERS))
#define TX_RING_MOD_MASK		(TX_RING_SIZE - 1)
#define TX_RING_LEN_BITS		((PCNET32_LOG_TX_BUFFERS) << 12)

#define RX_RING_SIZE			(1 << (PCNET32_LOG_RX_BUFFERS))
#define RX_RING_MOD_MASK		(RX_RING_SIZE - 1)
#define RX_RING_LEN_BITS		((PCNET32_LOG_RX_BUFFERS) << 4)

#define PKT_BUF_SZ		1544

/* Offsets from base I/O address. */
#define PCNET32_WIO_RDP		0x10
#define PCNET32_WIO_RAP		0x12
#define PCNET32_WIO_RESET	0x14
#define PCNET32_WIO_BDP		0x16

#define PCNET32_DWIO_RDP	0x10
#define PCNET32_DWIO_RAP	0x14
#define PCNET32_DWIO_RESET	0x18
#define PCNET32_DWIO_BDP	0x1C

#define PCNET32_TOTAL_SIZE 0x20

/* some PCI ids */
#ifndef PCI_DEVICE_ID_AMD_LANCE
#define PCI_VENDOR_ID_AMD	      0x1022
#define PCI_DEVICE_ID_AMD_LANCE	      0x2000
#endif
#ifndef PCI_DEVICE_ID_AMD_PCNETHOME
#define PCI_DEVICE_ID_AMD_PCNETHOME   0x2001
#endif


#define CRC_POLYNOMIAL_LE 0xedb88320UL  /* Ethernet CRC, little endian */

/* The PCNET32 Rx and Tx ring descriptors. */
struct pcnet32_rx_head {
    u32 base;
    s16 buf_length;
    s16 status;    
    u32 msg_length;
    u32 reserved;
};
	
struct pcnet32_tx_head {
    u32 base;
    s16 length;
    s16 status;
    u32 misc;
    u32 reserved;
};

/* The PCNET32 32-Bit initialization block, described in databook. */
struct pcnet32_init_block {
    u16 mode;
    u16 tlen_rlen;
    u8  phys_addr[6];
    u16 reserved;
    u32 filter[2];
    /* Receive and transmit ring base, along with extra bits. */    
    u32 rx_ring;
    u32 tx_ring;
};

/* PCnet32 access functions */
struct pcnet32_access {
    u16 (*read_csr)(unsigned long, int);
    void (*write_csr)(unsigned long, int, u16);
    u16 (*read_bcr)(unsigned long, int);
    void (*write_bcr)(unsigned long, int, u16);
    u16 (*read_rap)(unsigned long);
    void (*write_rap)(unsigned long, u16);
    void (*reset)(unsigned long);
#ifdef HOMENET
    u16 (*read_mii)(unsigned long, int);
    void (*write_mii)(unsigned long, int, u16);
#endif
};

struct pcnet32_private {
    /* The Tx and Rx ring entries must be aligned on 16-byte boundaries in 32bit mode. */
    struct pcnet32_rx_head   rx_ring[RX_RING_SIZE];
    struct pcnet32_tx_head   tx_ring[TX_RING_SIZE];
    struct pcnet32_init_block	init_block;
    const char *name;
    /* The saved address of a sent-in-place packet/buffer, for skfree(). */
    struct sk_buff *tx_skbuff[TX_RING_SIZE];
    struct sk_buff *rx_skbuff[RX_RING_SIZE];
    struct pcnet32_access a;
    void *origmem;
    int cur_rx, cur_tx;			/* The next free ring entry */
    int dirty_rx, dirty_tx;		        /* The ring entries to be free()ed. */
    struct net_device_stats stats;
    char tx_full;
    int  options;
    int  shared_irq:1,                      /* shared irq possible */
         full_duplex:1,                     /* full duplex possible */
         mii:1;                             /* mii port available */
#ifdef MODULE
    struct device *next;
#endif    
#ifdef HOMENET
    struct timer_list timer;                /* To monitor PHY changes and  */
                                            /* network power/speed changes */
    int             chip_version;
    int             device_type;
    int             cable_disconnected;
    u32             reset_phy_status;
    u32             number_phys;
    struct phy_desc phy_list[MAX_PHYS];     /* to store info about phys */
	u16			    active_phy;
	u16			    phy_search_sequence[MAX_PHYS];
	u16             level_ctrl;          // HomeRun Phy Reg 30
	u16             analog_test;         // HomeRun Phy Reg 31
	u32             home_run0;              // HomeRun Phy Register programming
										 // High order 2 bytes : Register number
										 // Low order 2 bytes : Register value
	u32             home_run1;
	u32             ipg;
	u32             sync_rate;
	u32             home_run_state;
	u32             phy_selected;
	struct auto_config HLAC;
	u32			    auto_neg_settle_count;
	u32             current_speed_and_power;
#endif
};

int  pcnet32_probe(struct device *);
static int  pcnet32_probe1(struct device *, unsigned long, unsigned char, int, int);
static int  pcnet32_open(struct device *);
static int  pcnet32_init_ring(struct device *);
static int  pcnet32_start_xmit(struct sk_buff *, struct device *);
static int  pcnet32_rx(struct device *);
static void pcnet32_interrupt(int, void *, struct pt_regs *);
static int  pcnet32_close(struct device *);
static struct net_device_stats *pcnet32_get_stats(struct device *);
static void pcnet32_set_multicast_list(struct device *);
#ifdef HAVE_PRIVATE_IOCTL
static int  pcnet32_mii_ioctl(struct device *, struct ifreq *, int);
#endif


struct pcnet32_pci_id_info {
    const char *name;
    u16 vendor_id, device_id, svid, sdid, flags;
    int io_size;
    int (*probe1) (struct device *, unsigned long, unsigned char, int, int);
};



